import numpy as np
import scipy.interpolate
import scipy.integrate
import disclap
from gausslin import GaussLinearization
import functools

# CONVENTIONS

# Infinitesimal invariants of the affine sphere are passed around as "a-vectors",
# which are 5-tuples consisting of
#   (u(p), du/dx(p), du/dy(p), Re(C(p)), Im(C(p)))

# The frame field is considered as a 4x4 matrix in which the ROWS are
#   sigma_1, sigma_2, N, f

def coefmat_dx(a):
    u = a[0]
    ux = a[1]
    uy = a[2]
    cr = a[3]
    ci = a[4]

    emu = np.exp(-u)
    eu = np.exp(u)

    return np.matrix([ [0.0, -uy, emu*np.sqrt(2)*cr, np.sqrt(2)*eu],
                       [uy, 0.0, -emu*np.sqrt(2)*ci, 0.0],
                       [emu*np.sqrt(2)*cr, -emu*np.sqrt(2)*ci, 0.0, 0.0],
                       [np.sqrt(2)*eu, 0.0, 0.0, 0.0] ])
 
def coefmat_dy(a):
    u = a[0]
    ux = a[1]
    uy = a[2]
    cr = a[3]
    ci = a[4]

    emu = np.exp(-u)
    eu = np.exp(u)

    return np.matrix([ [0.0, ux, -emu*np.sqrt(2)*ci, 0.0],
                       [-ux, 0.0, -emu*np.sqrt(2)*cr, np.sqrt(2)*eu],
                       [-emu*np.sqrt(2)*ci, -emu*np.sqrt(2)*cr, 0.0, 0.0],
                       [0.0, np.sqrt(2)*eu, 0.0, 0.0] ] )


def coefmat(theta,a):
    return np.cos(theta)*coefmat_dx(a) + np.sin(theta)*coefmat_dy(a)

def odeA(theta,pcdata,t,y):
    m = coefmat(theta,pcdata(t * np.exp(1j*theta)))
    ya = np.matrix(y)
    ya.shape = (4,4)
    return (m * ya).ravel() # flatten 3x3 matrix to 9-vector

def phase(z):
    a = np.angle(z)
    if a < 0:
        return 2.0*np.pi+a
    else:
        return a

class InducedMetric(GaussLinearization):
    DEFAULT_RET_TYPE='frame'
    def __init__(self,*args,**kwargs):
        GaussLinearization.__init__(self,*args,**kwargs)

        # Prepare for ODE integration
        self.ux = disclap.discdx(self.grid) * self.u
        self.uy = disclap.discdy(self.grid) * self.u
        self.yinit = np.eye(4).ravel()
        
        self.uinterp = scipy.interpolate.RectBivariateSpline(self.grid.x,self.grid.y,np.transpose(self.u.reshape((self.grid.nx,self.grid.ny))))
        self.uxinterp = scipy.interpolate.RectBivariateSpline(self.grid.x,self.grid.y,np.transpose(self.ux.reshape((self.grid.nx,self.grid.ny))))
        self.uyinterp = scipy.interpolate.RectBivariateSpline(self.grid.x,self.grid.y,np.transpose(self.uy.reshape((self.grid.nx,self.grid.ny))))

    def pcdata(self,z):
        cval = self.c(z)
        return np.array( [self.uinterp(z.real,z.imag),
                          self.uxinterp(z.real,z.imag),
                          self.uyinterp(z.real,z.imag),
                          cval.real,
                          cval.imag] )

    def prepare_output(self,F,return_type='frame'):
        if return_type == 'affine':
            return F[3,1:] / F[3,0]
        elif return_type == 'homog':
            return F[3,:]
        elif return_type == 'frame':
            return F

    def integrate_polar(self,r=None,theta=0.0,step=0.01,tol=0.00001,return_type='frame'):
        # If r not given, use circle 90% of radius of grid incircle
        if r==None:
            r = 0.9 * self.grid.r
        odef = functools.partial(odeA,theta,self.pcdata)
        solver = scipy.integrate.ode(odef)
        solver.set_integrator("vode",method="adams",with_jacobian=False,first_step=(step*r),rtol=tol,nsteps=50000)
        solver.set_initial_value(self.yinit,0.0)
        solver.integrate(r)
        return self.prepare_output(solver.y.reshape((4,4)), return_type)

    def integrate_point(self,z,step=0.01,tol=0.00001,return_type=DEFAULT_RET_TYPE):
        return self.integrate_polar(r=abs(z),theta=np.angle(z),step=step,tol=tol,return_type=return_type)

    def integrate_rays(self,n,r=None,theta0=0.0,step=0.01,tol=0.0001,return_type=DEFAULT_RET_TYPE):
        thetalist = np.linspace(theta0, theta0 + 2*np.pi, num=n, endpoint=False)
        return [ self.integrate_polar(r=r,theta=theta,step=step,tol=tol,return_type=return_type) for theta in thetalist ]

    def integrate_points(self,zlist,step=0.01,tol=0.00001,return_type=DEFAULT_RET_TYPE):
        return [ self.integrate_point(z,step=step,tol=tol,return_type=return_type) for z in zlist ]

def _moduletest():
    import squaregrid
    def c(z):
        return z*z
    gr = squaregrid.SquareGrid(5.0,100)
    B = InducedMetric(c,gr)
    vlist = B.integrate_rays(50)
    for v in vlist:
        print(v[0],v[1])
        
if __name__=='__main__':
    _moduletest()
