import numpy as np
from squaregrid import SquareGrid
from inducedmet import InducedMetric
from functools import partial
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
from scipy import interpolate
from drawpoly import *
import cv2
import glob
import os
import HQD
import random

def poly(theta,coefs,z):
    result = 0
    for c in reversed(coefs):
        result = result*z + c
    return np.exp(1j * theta)*result

def find_vertices(coefs):
    """
    Input: coefficients of a monic and centered polynomial as a list
    Output: list of vectors in R^4 representing the vertices of the polygon
    """
    theta = 0
    gr = SquareGrid(5.0,100)
    c = partial(poly,theta,coefs)
    met = InducedMetric(c,gr)
    degree = len(coefs)-1
    star_angle = -np.angle(coefs[-1])/float(degree+2)
    nvert = 2*degree+4
    framelist = met.integrate_rays(n=nvert,theta0=star_angle,r=6,step=0.005,tol=0.00001,return_type='frame')
    vertices=[]
    for F in framelist:
        w,v = np.linalg.eig(np.transpose(F))
        i = np.argmax(abs(w))
        vertices.append(v[:,i])
    vertices.append(vertices[0])
    return vertices

def draw_polygon(coefs):
    """
    Input: coefficients of a monic and centered polynomial
    Output: image of the polygon as object
    """
    #HQD.change_HQD(coefs)
    vert = find_vertices(coefs)
    coord = scaling(vert)
    fig = show_outline(coord)
    return fig

def draw_surface(coefs):
    """
    Input: coefficients of a polynomial
    Output: image of the surface as object
    """
    HQD.change_HQD(coefs)
    # Integrating the ODE in a grid
    theta = 0
    gr = SquareGrid(5.0,100)
    c = partial(poly,theta,coefs)
    met = InducedMetric(c,gr)
    parameter = np.linspace(-3, 3, 100)
    X, Y = np.meshgrid(parameter, parameter)
    X = X.ravel()
    Y = Y.ravel()
    Z = X + 1j*Y
    framelist = met.integrate_points(Z,step=0.005,tol=0.000001,return_type='frame') #other lines are time expensive but this is by far the most expensive
    ptlist = [] 
    for F in framelist: 
        ptlist.append(F[3,:-1]/F[3,3])
    
    # Plotting the surface 
    x = []
    y = []
    z = []

    for idx in range(len(ptlist)):
        x.append(ptlist[idx][0])
        y.append(ptlist[idx][1])
        z.append(ptlist[idx][2])

    x = np.array(x)
    y = np.array(y)
    z = np.array(z)

    fig = plt.figure(figsize=(15, 15))
    ax = plt.axes(projection = '3d')
    ax.plot_surface(x.reshape(100,100), y.reshape(100,100), z.reshape(100,100), cmap='summer')
    
    # Plotting AdS space
    angle = np.linspace(0, 2 * np.pi, 50)
    time = np.linspace(-3, 3, 3000)
    t, theta = np.meshgrid(time, angle)
    X2 = np.sqrt(1 + t ** 2) * np.cos(theta)
    Y2 = np.sqrt(1 + t ** 2) * np.sin(theta)
    Z2 = t
    ax.plot_surface(X2, Y2, Z2, color='grey', alpha=0.2)
    ax.view_init(30, 150)
    
    return fig

def create_video(polynomials, surface=False, filename='polyplot'):
    """
    Input:
        polynomials: a list of polynomials with complex coefficients, stored as lists of coefficients
        surface: a boolean which signals whether we make images of surfaces or polygons
        filename: name of video file
    Output: a video out of a sequence of surfaces (if surfaces is true), which are given by the polynomials in the list. if surface
            is false, the video is of polygons in Ads
    """
    img_array = []
    if surface:
        for idx in range(len(polynomials)):
                fig = draw_surface(polynomials[idx])
                plt.savefig('surface.png')
                img = cv2.imread('surface.png')
                height, width, layers = img.shape
                size = (width,height)
                img_array.append(img)
    else:
        for idx in range(len(polynomials)):
            fig = draw_polygon(polynomials[idx])
            plt.savefig('surface.png')
            img = cv2.imread('surface.png')
            height, width, layers = img.shape
            size = (width,height)
            img_array.append(img)
    
    out = cv2.VideoWriter(filename + '.avi',cv2.VideoWriter_fourcc(*'DIVX'), 5, size)
     
    for i in range(len(img_array)):
        out.write(img_array[i])
    
    os.remove('surface.png')
    out.release()


def create_scaling_video(coefs, scales, surface=False, filename='polyplot'):
    """
    Input:
        coefs: the coefficients of a polynomial with complex coefficients
        scales: a list of complex numbers to scale coefs by
        surface: true if plotting surfaces, false if plotting polygons
        filename: name of video file
    Output: a video of the surfaces/polygons of the polynomial coefs times each
            scalar in scales
    """
    polynomials = []
    for scalar in scales:
        new_poly = [x * scalar for x in coefs]
        polynomials.append(new_poly)
    
    create_video(polynomials, surface, filename)
    
def create_circle_video(coefs, surface=False, filename='polycircle'):
    """
    Input:
        coefs: the coefficients of a polynomial with complex coefficients
        surface: true if plotting surfaces, false if plotting polygons
        filename: name of video file
    Output: a video of the surfaces/polygons of the rotated polynomial
    """
    degree = len(coefs)-1
    phases = np.linspace(0, degree+2, (degree+2)*25)
    polynomials = []
    for phase in phases:
        new_poly = []
        for k in range(len(coefs)):
            new_poly.append(coefs[k]*np.exp(2*np.pi*1j*phase*(degree-k)/(degree+2)))
        polynomials.append(new_poly)
        
    create_video(polynomials, surface, filename)
#scales = np.linspace(1, 20, 20)
#create_scaling_video([random.random(), random.random(), random.random(), random.random()], scales)